from dataclasses import dataclass
from decimal import Decimal
from enum import Enum


# cf. https://developer.oanda.com/rest-live-v20/order-df/#TimeInForce
class TimeInForce(Enum):
    GTC = "GTC"  # Good until Cancelled
    GTD = "GTD"  # Good until Date
    GFD = "GFD"  # Good For Day
    FOK = "FOK"  # Filled Or Killed
    IOC = "IOC"  # Immediately partially filled Or Cancelled


# cf. https://developer.oanda.com/rest-live-v20/order-df/#OrderTriggerCondition
class OrderTriggerCondition(Enum):
    DEFAULT = "DEFAULT"
    # ...


# cf. https://developer.oanda.com/rest-live-v20/order-df/#OrderPositionFill
class OrderPositionFill(Enum):
    OPEN_ONLY = "OPEN_ONLY"
    REDUCE_FIRST = "REDUCE_FIRST"
    REDUCE_ONLY = "REDUCE_ONLY"
    DEFAULT = "DEFAULT"


# cf. https://developer.oanda.com/rest-live-v20/primitives-df/#HomeConversionFactors
@dataclass
class HomeConversionFactors:
    gain_quote_home: Decimal
    loss_quote_home: Decimal
    gain_base_home: Decimal
    loss_base_home: Decimal


def parse_home_conversion_factors(data: dict) -> HomeConversionFactors:
    return HomeConversionFactors(
        gain_quote_home=data["gainQuoteHome"],
        loss_quote_home=data["lossQuoteHome"],
        gain_base_home=data["gainBaseHome"],
        loss_base_home=data["lossBaseHome"],
    )



from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Optional

from .common import OrderPositionFill, OrderTriggerCondition, TimeInForce
from .transaction import (
    ClientExtensions,
    LimitOrderTransaction,
    MarketOrderTransaction,
    OrderCancelTransaction,
    OrderFillTransaction,
    StopLossDetails,
    TakeProfitDetails,
    parse_limit_order_transaction,
    parse_market_order_transaction,
    parse_order_cancel_transaction,
    parse_order_fill_transaction,
)


# cf. https://developer.oanda.com/rest-live-v20/order-df/#OrderType
class OrderType(Enum):
    MARKET = "MARKET"
    LIMIT = "LIMIT"
    STOP = "STOP"
    MARKET_IF_TOUCHED = "MARKET_IF_TOUCHED"
    TAKE_PROFIT = "TAKE_PROFIT"
    STOP_LOSS = "STOP_LOSS"
    GUARANTEED_STOP_LOSS = "GUARANTEED_STOP_LOSS"
    TRAILING_STOP_LOSS = "TRAILING_STOP_LOSS"
    FIXED_PRICE = "FIXED_PRICE"


# cf. https://developer.oanda.com/rest-live-v20/order-df/#OrderState
class OrderState(Enum):
    PENDING = "PENDING"
    FILLED = "FILLED"
    TRIGGERED = "TRIGGERED"
    CANCELLED = "CANCELLED"


# cf. https://developer.oanda.com/rest-live-v20/order-df/#MarketOrderRequest
@dataclass
class MarketOrderRequest:
    instrument: str
    units: Decimal
    type: OrderType = OrderType.MARKET
    time_in_force: TimeInForce = TimeInForce.FOK
    price_bound: Optional[Decimal] = None
    position_fill: OrderPositionFill = OrderPositionFill.DEFAULT
    client_extensions: Optional[ClientExtensions] = None
    take_profit_on_fill: Optional[TakeProfitDetails] = None
    stop_loss_on_fill: Optional[StopLossDetails] = None
    # guaranteed_stop_loss_on_fill: GuaranteedStopLossDetails
    # trailing_stop_loss_on_fill: TrailingStopLossDetails
    trade_client_extensions: Optional[ClientExtensions] = None


# cf. https://developer.oanda.com/rest-live-v20/order-df/#LimitOrderRequest
@dataclass
class LimitOrderRequest:
    instrument: str
    units: Decimal
    price: Decimal
    type: OrderType = OrderType.LIMIT
    time_in_force: TimeInForce = TimeInForce.GTC
    gtd_time: Optional[datetime] = None
    position_fill: OrderPositionFill = OrderPositionFill.DEFAULT
    trigger_condition: OrderTriggerCondition = OrderTriggerCondition.DEFAULT
    client_extensions: Optional[ClientExtensions] = None
    take_profit_on_fill: Optional[TakeProfitDetails] = None
    stop_loss_on_fill: Optional[StopLossDetails] = None
    # guaranteed_stop_loss_on_fill: GuaranteedStopLossDetails
    # trailing_stop_loss_on_fill: TrailingStopLossDetails
    trade_client_extensions: Optional[ClientExtensions] = None


# cf. https://developer.oanda.com/rest-live-v20/order-ep/
@dataclass
class CreateOrderResponse:
    related_transaction_ids: list[str]
    last_transaction_id: str


# cf. https://developer.oanda.com/rest-live-v20/order-ep/
@dataclass
class CreateMarketOrderResponse(CreateOrderResponse):
    order_create_transaction: MarketOrderTransaction
    order_fill_transaction: OrderFillTransaction


def parse_create_market_order_response(data: dict) -> CreateMarketOrderResponse:
    return CreateMarketOrderResponse(
        order_create_transaction=parse_market_order_transaction(data["orderCreateTransaction"]),
        order_fill_transaction=parse_order_fill_transaction(data["orderFillTransaction"]),
        related_transaction_ids=data["relatedTransactionIDs"],
        last_transaction_id=data["lastTransactionID"],
    )


# cf. https://developer.oanda.com/rest-live-v20/order-ep/
@dataclass
class CreateLimitOrderResponse(CreateOrderResponse):
    order_create_transaction: LimitOrderTransaction


def parse_create_limit_order_response(data: dict) -> CreateLimitOrderResponse:
    return CreateLimitOrderResponse(
        order_create_transaction=parse_limit_order_transaction(data["orderCreateTransaction"]),
        related_transaction_ids=data["relatedTransactionIDs"],
        last_transaction_id=data["lastTransactionID"],
    )


# cf. https://developer.oanda.com/rest-live-v20/order-ep/
@dataclass
class CancelOrderResponse(CreateOrderResponse):
    order_cancel_transaction: OrderCancelTransaction


def parse_cancel_order_response(data: dict) -> CancelOrderResponse:
    return CancelOrderResponse(
        order_cancel_transaction=parse_order_cancel_transaction(data["orderCancelTransaction"]),
        related_transaction_ids=data["relatedTransactionIDs"],
        last_transaction_id=data["lastTransactionID"],
    )



from .datetime import format_datetime as format_datetime
from .datetime import parse_time as parse_time
from .json import JSONEncoder as JSONEncoder
from .json import remove_none as remove_none
from .json import to_camel_case as to_camel_case



from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from typing import Optional

from ..helper import parse_time


# https://developer.oanda.com/rest-live-v20/pricing-common-df/#PriceBucket
@dataclass
class PriceBucket:
    price: Decimal
    liquidity: int


def parse_price_bucket(data: dict) -> PriceBucket:
    return PriceBucket(
        price=Decimal(data["price"]),
        liquidity=data["liquidity"],
    )


# https://developer.oanda.com/rest-live-v20/pricing-df/#ClientPrice
@dataclass
class ClientPrice:
    type: str
    instrument: Optional[str]
    # BUG: pricing stream API では time が返されるが、
    # transaction stream API の子要素では timestamp が返される.
    time: Optional[datetime]
    timestamp: Optional[datetime]
    tradeable: Optional[bool]
    bids: list[PriceBucket]
    asks: list[PriceBucket]
    closeout_bid: Decimal
    closeout_ask: Decimal


def parse_client_price(data: dict) -> ClientPrice:
    return ClientPrice(
        type=data["type"] if "type" in data else "PRICE",
        instrument=data["instrument"] if "instrument" in data else None,
        time=parse_time(data["time"]) if "time" in data else None,
        timestamp=parse_time(data["timestamp"]) if "timestamp" in data else None,
        tradeable=data["tradeable"] if "tradeable" in data else None,
        bids=[parse_price_bucket(x) for x in data["bids"]],
        asks=[parse_price_bucket(x) for x in data["asks"]],
        closeout_bid=Decimal(data["closeoutBid"]),
        closeout_ask=Decimal(data["closeoutAsk"]),
    )


# https://developer.oanda.com/rest-live-v20/pricing-df/#PricingHeartbeat
@dataclass
class PricingHeartbeat:
    type: str
    time: datetime



import os
import threading
import warnings
from pathlib import Path

import yaml


class OANDAConfig:
    def __init__(self):
        self._lock = threading.Lock()
        self._initialized = False
        self.rest_url = None
        self.streaming_url = None
        self.account = None
        self.token = None

    def configure(
        self,
        *,
        rest_url=None,
        streaming_url=None,
        account=None,
        token=None,
        force=False,
    ):
        with self._lock:
            if self._initialized and not force:
                warnings.warn(
                    "APIConfig is already configured. Use force=True to override.",
                    stacklevel=2,
                )
                return

            self.rest_url = rest_url
            self.streaming_url = streaming_url
            self.account = account
            self.token = token
            self._initialized = True

    def load_from_env(self):
        self.configure(
            rest_url=os.getenv("OANDA_REST_URL"),
            streaming_url=os.getenv("OANDA_STREAMING_URL"),
            account=os.getenv("OANDA_ACCOUNT"),
            token=os.getenv("OANDA_TOKEN"),
        )

    def load_from_file(self, path=None):
        if path is None:
            path = Path.home() / ".strats_oanda.yaml"

        if not Path(path).exists():
            warnings.warn(f"Config file not found: {path}", stacklevel=2)
            return

        with open(path) as f:
            data = yaml.safe_load(f)

        self.configure(
            rest_url=data.get("rest_url"),
            streaming_url=data.get("streaming_url"),
            account=data.get("account"),
            token=data.get("token"),
        )

    def is_configured(self):
        return self._initialized

    @property
    def account_rest_url(self) -> str:
        return f"{self.rest_url}/v3/accounts/{self.account}"

    @property
    def account_streaming_url(self) -> str:
        return f"{self.streaming_url}/v3/accounts/{self.account}"


_config = OANDAConfig()


def basic_config(
    *,
    rest_url=None,
    streaming_url=None,
    account=None,
    token=None,
    force=False,
    use_env=False,
    use_file=False,
    file_path=None,
):
    if use_env:
        _config.load_from_env()
    elif use_file:
        _config.load_from_file(file_path)
    else:
        _config.configure(
            rest_url=rest_url,
            streaming_url=streaming_url,
            account=account,
            token=token,
            force=force,
        )


def get_config() -> OANDAConfig:
    return _config



from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Optional

from ..helper import parse_time
from .common import (
    HomeConversionFactors,
    OrderPositionFill,
    OrderTriggerCondition,
    TimeInForce,
    parse_home_conversion_factors,
)
from .pricing import ClientPrice, parse_client_price


# cf. https://developer.oanda.com/rest-live-v20/transaction-df/#ClientExtensions
@dataclass
class ClientExtensions:
    id: str
    tag: str
    comment: str


# cf. https://developer.oanda.com/rest-live-v20/transaction-df/#TakeProfitDetails
@dataclass
class TakeProfitDetails:
    price: Decimal
    time_in_force: TimeInForce = TimeInForce.GTC
    gtd_time: Optional[datetime] = None
    client_extensions: Optional[ClientExtensions] = None


# cf. https://developer.oanda.com/rest-live-v20/transaction-df/#StopLossDetails
@dataclass
class StopLossDetails:
    price: Optional[Decimal] = None
    distance: Optional[Decimal] = None
    time_in_force: TimeInForce = TimeInForce.GTC
    gtd_time: Optional[datetime] = None
    client_extensions: Optional[ClientExtensions] = None


# https://developer.oanda.com/rest-live-v20/transaction-df/#MarketOrderReason
class MarketOrderReason(Enum):
    CLIENT_ORDER = "CLIENT_ORDER"
    TRADE_CLOSE = "TRADE_CLOSE"
    POSITION_CLOSEOUT = "POSITION_CLOSEOUT"
    MARGIN_CLOSEOUT = "MARGIN_CLOSEOUT"
    DELAYED_TRADE_CLOSEOUT = "DELAYED_TRADE_CLOSEOUT"


# https://developer.oanda.com/rest-live-v20/transaction-df/#LimitOrderReason
class LimitOrderReason(Enum):
    CLIENT_ORDER = "CLIENT_ORDER"
    REPLACEMENT = "REPLACEMENT"


# https://developer.oanda.com/rest-live-v20/transaction-df/#OrderCancelReason
class OrderCancelReason(Enum):
    CLIENT_REQUEST = "CLIENT_REQUEST"
    # ...


# https://developer.oanda.com/rest-live-v20/transaction-df/#OrderFillReason
class OrderFillReason(Enum):
    LIMIT_ORDER = "LIMIT_ORDER"
    STOP_ORDER = "STOP_ORDER"
    MARKET_IF_TOUCHED_ORDER = "MARKET_IF_TOUCHED_ORDER"
    TAKE_PROFIT_ORDER = "TAKE_PROFIT_ORDER"
    STOP_LOSS_ORDER = "STOP_LOSS_ORDER"
    GUARANTEED_STOP_LOSS_ORDER = "GUARANTEED_STOP_LOSS_ORDER"
    TRAILING_STOP_LOSS_ORDER = "TRAILING_STOP_LOSS_ORDER"
    MARKET_ORDER = "MARKET_ORDER"
    MARKET_ORDER_TRADE_CLOSE = "MARKET_ORDER_TRADE_CLOSE"
    MARKET_ORDER_POSITION_CLOSEOUT = "MARKET_ORDER_POSITION_CLOSEOUT"
    MARKET_ORDER_MARGIN_CLOSEOUT = "MARKET_ORDER_MARGIN_CLOSEOUT"
    MARKET_ORDER_DELAYED_TRADE_CLOSE = "MARKET_ORDER_DELAYED_TRADE_CLOSE"
    FIXED_PRICE_ORDER = "FIXED_PRICE_ORDER"
    FIXED_PRICE_ORDER_PLATFORM_ACCOUNT_MIGRATION = "FIXED_PRICE_ORDER_PLATFORM_ACCOUNT_MIGRATION"
    FIXED_PRICE_ORDER_DIVISION_ACCOUNT_MIGRATION = "FIXED_PRICE_ORDER_DIVISION_ACCOUNT_MIGRATION"
    FIXED_PRICE_ORDER_ADMINISTRATIVE_ACTION = "FIXED_PRICE_ORDER_ADMINISTRATIVE_ACTION"


# cf. https://developer.oanda.com/rest-live-v20/transaction-df/#MarketOrderTradeClose
@dataclass
class MarketOrderTradeClose:
    trade_id: str
    client_trade_id: str
    units: Decimal


# cf. https://developer.oanda.com/rest-live-v20/transaction-df/#MarketOrderPositionCloseout
@dataclass
class MarketOrderPositionCloseout:
    instrument: str
    units: Decimal


# https://developer.oanda.com/rest-live-v20/transaction-df/#TradeOpen
@dataclass
class TradeOpen:
    trade_id: str
    units: Decimal
    price: Decimal
    # guaranteedExecutionFee: Decimal
    # quoteGuaranteedExecutionFee: Decimal
    half_spread_cost: Decimal
    initial_margin_required: Decimal
    client_extensions: Optional[ClientExtensions] = None


def parse_trade_open(data: dict) -> TradeOpen:
    return TradeOpen(
        trade_id=data["tradeID"],
        units=Decimal(data["units"]),
        price=Decimal(data["price"]),
        half_spread_cost=Decimal(data["halfSpreadCost"]),
        initial_margin_required=Decimal(data["initialMarginRequired"]),
    )


# https://developer.oanda.com/rest-live-v20/transaction-df/#TradeReduce
@dataclass
class TradeReduce:
    trade_id: str
    units: Decimal
    price: Decimal
    # ...


def parse_trade_reduce(data: dict) -> TradeReduce:
    return TradeReduce(
        trade_id=data["tradeID"],
        units=Decimal(data["units"]),
        price=Decimal(data["price"]),
    )


# cf. https://developer.oanda.com/rest-live-v20/transaction-df/#OrderFillTransaction
@dataclass
class Transaction:
    id: str
    time: datetime
    user_id: int
    account_id: str
    batch_id: str
    type: str
    request_id: str  # allow empty string


# type = MARKET_ORDER
# cf. https://developer.oanda.com/rest-live-v20/transaction-df/#MarketOrderTransaction
@dataclass
class MarketOrderTransaction(Transaction):
    instrument: str
    units: Decimal
    time_in_force: TimeInForce
    reason: MarketOrderReason
    price_bound: Optional[Decimal] = None
    position_fill: OrderPositionFill = OrderPositionFill.DEFAULT
    trade_close: Optional[MarketOrderTradeClose] = None
    long_position_closeout: Optional[MarketOrderPositionCloseout] = None
    short_position_closeout: Optional[MarketOrderPositionCloseout] = None
    # margin_closeout: Optional[] = None
    # delayed_trade_close: Optional[] = None
    client_extensions: Optional[ClientExtensions] = None
    trade_client_extensions: Optional[ClientExtensions] = None
    # take_profit_on_fill
    # stop_loss_on_fill
    # trailing_stop_loss_on_fill
    # guarantee_stop_loss_on_fill


def parse_market_order_transaction(data: dict) -> MarketOrderTransaction:
    return MarketOrderTransaction(
        id=data["id"],
        time=parse_time(data["time"]),
        user_id=data["userID"],
        account_id=data["accountID"],
        batch_id=data["batchID"],
        request_id=data["requestID"],
        type=data["type"],
        instrument=data["instrument"],
        units=Decimal(data["units"]),
        time_in_force=TimeInForce(data["timeInForce"]),
        price_bound=Decimal(data["price_bound"]) if "price_bound" in data else None,
        reason=MarketOrderReason(data["reason"]),
    )


# type = LIMIT_ORDER
# cf. https://developer.oanda.com/rest-live-v20/transaction-df/#LimitOrderTransaction
@dataclass
class LimitOrderTransaction(Transaction):
    instrument: str
    units: Decimal
    price: Decimal
    time_in_force: TimeInForce
    gtd_time: Optional[datetime]
    trigger_condition: OrderTriggerCondition
    reason: LimitOrderReason


def parse_limit_order_transaction(data: dict) -> LimitOrderTransaction:
    return LimitOrderTransaction(
        id=data["id"],
        time=parse_time(data["time"]),
        user_id=data["userID"],
        account_id=data["accountID"],
        batch_id=data["batchID"],
        request_id=data["requestID"],
        type=data["type"],
        instrument=data["instrument"],
        units=Decimal(data["units"]),
        price=Decimal(data["price"]),
        time_in_force=TimeInForce(data["timeInForce"]),
        gtd_time=parse_time(data["gtdTime"]) if "gtdTime" in data else None,
        trigger_condition=OrderTriggerCondition(data["triggerCondition"]),
        reason=LimitOrderReason(data["reason"]),
    )


# cf. https://developer.oanda.com/rest-live-v20/transaction-df/#OrderCancelTransaction
@dataclass
class OrderCancelTransaction(Transaction):
    order_id: str
    reason: OrderCancelReason
    client_order_id: Optional[str] = None
    replaced_by_order_id: Optional[str] = None


def parse_order_cancel_transaction(data: dict) -> OrderCancelTransaction:
    return OrderCancelTransaction(
        id=data["id"],
        time=parse_time(data["time"]),
        user_id=data["userID"],
        account_id=data["accountID"],
        batch_id=data["batchID"],
        request_id=data["requestID"],
        type=data["type"],
        order_id=data["orderID"],
        reason=OrderCancelReason(data["reason"]),
    )


# cf. https://developer.oanda.com/rest-live-v20/transaction-df/#OrderFillTransaction
@dataclass
class OrderFillTransaction(Transaction):
    order_id: str
    client_order_id: Optional[str]
    instrument: str
    units: Decimal
    home_conversion_factors: HomeConversionFactors
    full_vwap: Decimal
    full_price: ClientPrice
    reason: OrderFillReason
    pl: Decimal
    quote_pl: Decimal
    financing: Decimal
    base_financing: Decimal
    quote_financing: Optional[Decimal]
    commission: Decimal
    guaranteed_execution_fee: Decimal
    quote_guaranteed_execution_fee: Decimal
    account_balance: Decimal
    trade_opened: Optional[TradeOpen]
    trades_closed: Optional[list[TradeReduce]]
    trade_reduced: Optional[TradeReduce]
    half_spread_cost: Decimal


def parse_order_fill_transaction(data: dict) -> OrderFillTransaction:
    return OrderFillTransaction(
        id=data["id"],
        account_id=data["accountID"],
        user_id=data["userID"],
        batch_id=data["batchID"],
        request_id=data["requestID"] if "requestID" in data else "",
        time=parse_time(data["time"]),
        type=data["type"],
        order_id=data["orderID"],
        client_order_id=data["clientOrderID"] if "clientOrderID" in data else None,
        instrument=data["instrument"],
        units=Decimal(data["units"]),
        home_conversion_factors=parse_home_conversion_factors(data["homeConversionFactors"]),
        full_vwap=Decimal(data["fullVWAP"]),
        full_price=parse_client_price(data["fullPrice"]),
        reason=OrderFillReason(data["reason"]),
        pl=Decimal(data["pl"]),
        quote_pl=Decimal(data["quotePL"]),
        financing=Decimal(data["financing"]),
        base_financing=Decimal(data["baseFinancing"]),
        quote_financing=Decimal(data["quoteFinancing"]) if "quoteFinancing" in data else None,
        commission=Decimal(data["commission"]),
        guaranteed_execution_fee=Decimal(data["guaranteedExecutionFee"]),
        quote_guaranteed_execution_fee=Decimal(data["quoteGuaranteedExecutionFee"]),
        account_balance=Decimal(data["accountBalance"]),
        trade_opened=parse_trade_open(data["tradeOpened"]) if "tradeOpened" in data else None,
        trades_closed=[parse_trade_reduce(x) for x in data["tradeClosed"]]
        if "tradeClosed" in data
        else None,
        trade_reduced=parse_trade_reduce(data["tradeReduced"]) if "tradeReduced" in data else None,
        half_spread_cost=Decimal(data["halfSpreadCost"]),
    )



from .common import HomeConversionFactors as HomeConversionFactors
from .common import OrderPositionFill as OrderPositionFill
from .common import OrderTriggerCondition as OrderTriggerCondition
from .common import TimeInForce as TimeInForce
from .instrument import Candlestick as Candlestick
from .instrument import CandlestickData as CandlestickData
from .instrument import CandlestickGranularity as CandlestickGranularity
from .order import CancelOrderResponse as CancelOrderResponse
from .order import CreateLimitOrderResponse as CreateLimitOrderResponse
from .order import CreateMarketOrderResponse as CreateMarketOrderResponse
from .order import LimitOrderRequest as LimitOrderRequest
from .order import MarketOrderRequest as MarketOrderRequest
from .order import OrderState as OrderState
from .order import OrderType as OrderType
from .order import parse_cancel_order_response as parse_cancel_order_response
from .order import parse_create_limit_order_response as parse_create_limit_order_response
from .order import parse_create_market_order_response as parse_create_market_order_response
from .pricing import ClientPrice as ClientPrice
from .pricing import PriceBucket as PriceBucket
from .pricing import PricingHeartbeat as PricingHeartbeat
from .pricing import parse_client_price as parse_client_price
from .pricing import parse_price_bucket as parse_price_bucket
from .transaction import ClientExtensions as ClientExtensions
from .transaction import LimitOrderReason as LimitOrderReason
from .transaction import LimitOrderTransaction as LimitOrderTransaction
from .transaction import MarketOrderReason as MarketOrderReason
from .transaction import MarketOrderTransaction as MarketOrderTransaction
from .transaction import OrderCancelReason as OrderCancelReason
from .transaction import OrderCancelTransaction as OrderCancelTransaction
from .transaction import OrderFillReason as OrderFillReason
from .transaction import OrderFillTransaction as OrderFillTransaction
from .transaction import StopLossDetails as StopLossDetails
from .transaction import TakeProfitDetails as TakeProfitDetails
from .transaction import TradeOpen as TradeOpen
from .transaction import TradeReduce as TradeReduce
from .transaction import Transaction as Transaction
from .transaction import parse_order_fill_transaction as parse_order_fill_transaction



from datetime import datetime


def format_datetime(t: datetime) -> str:
    """datetime(2024, 10, 14, 0, 0, 0)
    -> "2024-10-14T00:00:00.000000000Z"
    """
    return t.strftime("%Y-%m-%dT%H:%M:%S.%f") + "000Z"


def parse_time(s: str) -> datetime:
    """
    '2025-03-24T15:34:25.366624289Z'
    -> datetime(2025, 3, 24, 15, 34, 25)
    """
    if "." in s:
        datetime_part, frac_part = s.split(".")
        microsec_part = frac_part[:6]
        s = f"{datetime_part}.{microsec_part}Z"
    s = s.replace("Z", "+00:00")
    return datetime.fromisoformat(s)



import json
from decimal import Decimal
from enum import Enum

import inflection


class JSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, Enum):
            return o.value
        if isinstance(o, Decimal):
            return str(o)
        return super().default(o)


def remove_none(obj):
    """
    再帰的に None を含むキー (あるいは要素) を削除して返す。
    """
    if isinstance(obj, dict):
        return {k: remove_none(v) for k, v in obj.items() if v is not None}
    elif isinstance(obj, list):
        return [remove_none(v) for v in obj if v is not None]
    else:
        return obj


def to_camel_case(d):
    if isinstance(d, dict):
        return {inflection.camelize(k, False): to_camel_case(v) for k, v in d.items()}
    elif isinstance(d, list):
        return [to_camel_case(i) for i in d]
    else:
        return d



from .instrument import GetCandlesQueryParams as GetCandlesQueryParams
from .instrument import GetCandlesResponse as GetCandlesResponse
from .instrument import InstrumentClient as InstrumentClient
from .order import CancelOrderResponse as CancelOrderResponse
from .order import CreateLimitOrderResponse as CreateLimitOrderResponse
from .order import OrderClient as OrderClient
from .pricing import PricingStreamClient as PricingStreamClient
from .transaction import TransactionClient as TransactionClient



"""
Transaction Stream Endpoints
cf. https://developer.oanda.com/rest-live-v20/transaction-ep/
"""

import asyncio
import json
import logging
from collections.abc import AsyncGenerator
from typing import Optional

import aiohttp
from strats.exchange import StreamClient

from strats_oanda.config import get_config
from strats_oanda.model.transaction import (
    Transaction,
    parse_limit_order_transaction,
    parse_order_cancel_transaction,
    parse_order_fill_transaction,
)

logger = logging.getLogger(__name__)


class TransactionClient(StreamClient):
    def __init__(self):
        self.config = get_config()

    async def stream(self, stop_event: asyncio.Event) -> AsyncGenerator[Transaction]:
        url = f"{self.config.account_streaming_url}/transactions/stream"
        headers = {
            "Authorization": f"Bearer {self.config.token}",
        }

        async with aiohttp.ClientSession() as session:
            async with session.get(url, headers=headers) as resp:
                content_iter = resp.content.__aiter__()

                while not stop_event.is_set():
                    next_line_task = asyncio.create_task(content_iter.__anext__())
                    stop_task = asyncio.create_task(stop_event.wait())

                    done, pending = await asyncio.wait(
                        [next_line_task, stop_task],
                        return_when=asyncio.FIRST_COMPLETED,
                    )

                    # 不必要になった一時的な task は終了
                    for task in pending:
                        task.cancel()

                    if stop_task in done:
                        break

                    if next_line_task in done:
                        try:
                            line_bytes = next_line_task.result()
                        except (StopAsyncIteration, asyncio.CancelledError) as e:
                            logger.error(f"async iteration stopped or canceled. {e}")
                            break

                        line = line_bytes.decode("utf-8").strip()
                        if not line or "HEARTBEAT" in line:
                            continue

                        try:
                            data = json.loads(line)
                            tx_type = data.get("type")

                            tx: Optional[Transaction] = None
                            if tx_type == "LIMIT_ORDER":
                                tx = parse_limit_order_transaction(data)
                            elif tx_type == "ORDER_CANCEL":
                                tx = parse_order_cancel_transaction(data)
                            elif tx_type == "ORDER_FILL":
                                tx = parse_order_fill_transaction(data)
                            elif tx_type == "HEARTBEAT":
                                continue
                            else:
                                logger.warn(f"unknown transaction arrived. {data}")
                                continue

                            if tx is not None:
                                yield tx

                        except Exception as e:
                            logger.error(f"failed to parse message. err={e}, line={line}")
                            continue



from strats.model import PricesData

from ..model import ClientPrice


def client_price_to_prices_data(p: ClientPrice) -> PricesData:
    if len(p.bids) == 0 or len(p.asks) == 0:
        raise ValueError("bids and asks must be not empty")
    return PricesData(
        bid=p.bids[0].price,
        ask=p.asks[0].price,
    )



from datetime import datetime, timezone

from strats_oanda.helper import format_datetime, parse_time


def test_format_datetime():
    t = datetime(2024, 10, 14, 0, 0, 0)
    assert format_datetime(t) == "2024-10-14T00:00:00.000000000Z"


def test_parse_time():
    s = "2025-03-24T15:34:25.366624289Z"
    assert parse_time(s) == datetime(2025, 3, 24, 15, 34, 25, 366624, tzinfo=timezone.utc)



from datetime import datetime, timezone
from decimal import Decimal

from strats_oanda.model import (
    ClientPrice,
    HomeConversionFactors,
    OrderFillReason,
    OrderFillTransaction,
    PriceBucket,
    TradeOpen,
    parse_order_fill_transaction,
)


def test_parse_order_fill_transaction():
    data = {
        "accountBalance": "3000000.0110",
        "accountID": "101-009-31084545-001",
        "baseFinancing": "0",
        "batchID": "68",
        "commission": "0.0000",
        "financing": "0.0000",
        "fullPrice": {
            "asks": [{"liquidity": "250000", "price": "150.496"}],
            "bids": [{"liquidity": "250000", "price": "150.492"}],
            "closeoutAsk": "150.500",
            "closeoutBid": "150.488",
            "timestamp": "2025-03-26T13:50:53.182325174Z",
        },
        "fullVWAP": "150.492",
        "gainQuoteHomeConversionFactor": "1",
        "guaranteedExecutionFee": "0.0000",
        "halfSpreadCost": "0.0020",
        "homeConversionFactors": {
            "gainBaseHome": {"factor": "150.193012"},
            "gainQuoteHome": {"factor": "1"},
            "lossBaseHome": {"factor": "150.794988"},
            "lossQuoteHome": {"factor": "1"},
        },
        "id": "69",
        "instrument": "USD_JPY",
        "lossQuoteHomeConversionFactor": "1",
        "orderID": "68",
        "pl": "0.0000",
        "price": "150.492",
        "quoteGuaranteedExecutionFee": "0",
        "quotePL": "0",
        "reason": "MARKET_ORDER",
        "requestID": "79368253552993866",
        "requestedUnits": "-1",
        "time": "2025-03-26T13:50:53.280405861Z",
        "tradeOpened": {
            "guaranteedExecutionFee": "0.0000",
            "halfSpreadCost": "0.0020",
            "initialMarginRequired": "6.0198",
            "price": "150.492",
            "quoteGuaranteedExecutionFee": "0",
            "tradeID": "69",
            "units": "-1",
        },
        "type": "ORDER_FILL",
        "units": "-1",
        "userID": 31084545,
    }
    got = parse_order_fill_transaction(data)
    expect = OrderFillTransaction(
        id="69",
        time=datetime(2025, 3, 26, 13, 50, 53, 280405, tzinfo=timezone.utc),
        user_id=31084545,
        account_id="101-009-31084545-001",
        batch_id="68",
        type="ORDER_FILL",
        request_id="79368253552993866",
        order_id="68",
        client_order_id=None,
        instrument="USD_JPY",
        units=Decimal("-1"),
        home_conversion_factors=HomeConversionFactors(
            gain_quote_home={"factor": "1"},
            loss_quote_home={"factor": "1"},
            gain_base_home={"factor": "150.193012"},
            loss_base_home={"factor": "150.794988"},
        ),
        full_vwap=Decimal("150.492"),
        full_price=ClientPrice(
            type="PRICE",
            instrument=None,
            time=None,
            timestamp=datetime(2025, 3, 26, 13, 50, 53, 182325, tzinfo=timezone.utc),
            tradeable=None,
            bids=[
                PriceBucket(
                    price=Decimal("150.492"),
                    liquidity="250000",
                ),
            ],
            asks=[
                PriceBucket(
                    price=Decimal("150.496"),
                    liquidity="250000",
                ),
            ],
            closeout_bid=Decimal("150.488"),
            closeout_ask=Decimal("150.500"),
        ),
        reason=OrderFillReason.MARKET_ORDER,
        pl=Decimal("0.0000"),
        quote_pl=Decimal("0"),
        financing=Decimal("0.0000"),
        base_financing=Decimal("0"),
        quote_financing=None,
        commission=Decimal("0.0000"),
        guaranteed_execution_fee=Decimal("0.0000"),
        quote_guaranteed_execution_fee=Decimal("0"),
        account_balance=Decimal("3000000.0110"),
        trade_opened=TradeOpen(
            trade_id="69",
            units=Decimal("-1"),
            price=Decimal("150.492"),
            half_spread_cost=Decimal("0.0020"),
            initial_margin_required=Decimal("6.0198"),
            client_extensions=None,
        ),
        trades_closed=None,
        trade_reduced=None,
        half_spread_cost=Decimal("0.0020"),
    )
    assert got == expect



# Instrument Endpoint
# cf. https://developer.oanda.com/rest-live-v20/instrument-ep/
from dataclasses import dataclass
from datetime import datetime
from typing import Optional

import requests

from strats_oanda.helper import format_datetime
from strats_oanda.logger import logger
from strats_oanda.model.instrument import (
    Candlestick,
    CandlestickGranularity,
    parse_candlestick,
)


@dataclass
class GetCandlesQueryParams:
    count: Optional[int] = None
    from_time: Optional[datetime] = None
    to_time: Optional[datetime] = None


@dataclass
class GetCandlesResponse:
    instrument: str
    granularity: CandlestickGranularity
    candles: list[Candlestick]


def parse_get_candles_response(data) -> GetCandlesResponse:
    return GetCandlesResponse(
        instrument=data["instrument"],
        granularity=CandlestickGranularity(data["granularity"]),
        candles=[parse_candlestick(x) for x in data["candles"]],
    )


class InstrumentClient:
    def __init__(self, url, token):
        self.url = url
        self.token = token

    def get_candles(
        self,
        instrument: str,
        params: GetCandlesQueryParams,
    ) -> Optional[GetCandlesResponse]:
        url = f"{self.url}/v3/instruments/{instrument}/candles"
        payload = {
            # PricingComponent
            # Can contain any combination of the characters “M” (midpoint candles)
            # “B” (bid candles) and “A” (ask candles).
            # cf. https://developer.oanda.com/rest-live-v20/primitives-df/#PricingComponent
            "price": "M",
            "granularity": "M1",
        }
        if params.count is not None:
            payload["count"] = str(params.count)
        if params.from_time is not None:
            payload["from"] = format_datetime(params.from_time)
        if params.to_time is not None:
            payload["to"] = format_datetime(params.to_time)

        headers = {
            "Authorization": f"Bearer {self.token}",
            "Content-Type": "application/json",
        }
        res = requests.get(url, headers=headers, params=payload)

        if res.status_code == 200:
            return parse_get_candles_response(res.json())
        logger.error(f"Error get candles data: {res.status_code} {res.text}")
        return None



"""
Order Endpoints Client
cf. https://developer.oanda.com/rest-live-v20/order-ep/
"""

import asyncio
import json
import logging
from dataclasses import asdict
from typing import Optional

import aiohttp

from strats_oanda.config import get_config
from strats_oanda.helper import JSONEncoder, remove_none, to_camel_case
from strats_oanda.model import (
    CancelOrderResponse,
    CreateLimitOrderResponse,
    CreateMarketOrderResponse,
    LimitOrderRequest,
    MarketOrderRequest,
    parse_cancel_order_response,
    parse_create_limit_order_response,
    parse_create_market_order_response,
)

logger = logging.getLogger(__name__)


class OrderClient:
    def __init__(self, keepalive_timeout: float = 60.0, max_retries: int = 2):
        self.config = get_config()
        self.headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.config.token}",
        }
        self.keepalive_timeout = keepalive_timeout
        self.max_retries = max_retries
        self.session: Optional[aiohttp.ClientSession] = None
        self._owns_session = False

    async def open(self):
        if self.session is None or self.session.closed:
            connector = aiohttp.TCPConnector(keepalive_timeout=self.keepalive_timeout)
            self.session = aiohttp.ClientSession(headers=self.headers, connector=connector)
            self._owns_session = True

    async def close(self):
        if self._owns_session and self.session and not self.session.closed:
            await self.session.close()

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def _request_with_retry(
        self, method: str, url: str, **kwargs
    ) -> Optional[aiohttp.ClientResponse]:
        if self.session is None or self.session.closed:
            raise RuntimeError("ClientSession is not open. Call 'await open()' first.")

        for attempt in range(1, self.max_retries + 2):
            try:
                async with self.session.request(method, url, **kwargs) as res:
                    return res
            except (aiohttp.ClientError, asyncio.TimeoutError) as e:
                logger.warning(f"Request {method} {url} failed (attempt {attempt}): {e}")
                if attempt < self.max_retries + 1:
                    await asyncio.sleep(1)
                else:
                    logger.error("All retry attempts failed.")
                    return None
        return None  # ここには到達しない

    async def create_market_order(
        self,
        market_order: MarketOrderRequest,
    ) -> Optional[CreateMarketOrderResponse]:
        url = f"{self.config.account_rest_url}/orders"
        req = to_camel_case(remove_none({"order": asdict(market_order)}))
        order_data = json.dumps(req, cls=JSONEncoder)
        logger.info(f"create market order: {order_data}")

        res = await self._request_with_retry("POST", url, data=order_data)
        if res is None:
            return None

        if res.status == 201:
            data = await res.json()
            logger.info("create market order success")
            try:
                return parse_create_market_order_response(data)
            except Exception as e:
                logger.exception(f"Failed to parse market order response: {e}")
                return None
        else:
            text = await res.text()
            logger.error(f"error creating market order: {res.status} {text}")
            return None

    async def create_limit_order(
        self,
        limit_order: LimitOrderRequest,
    ) -> Optional[CreateLimitOrderResponse]:
        url = f"{self.config.account_rest_url}/orders"
        req = to_camel_case(remove_none({"order": asdict(limit_order)}))
        order_data = json.dumps(req, cls=JSONEncoder)
        logger.info(f"create limit order: {order_data}")

        res = await self._request_with_retry("POST", url, data=order_data)
        if res is None:
            return None

        if res.status == 201:
            data = await res.json()
            logger.info("create limit order success")
            return parse_create_limit_order_response(data)
        else:
            text = await res.text()
            logger.error(f"error creating limit order: {res.status} {text}")
            return None

    async def cancel_limit_order(self, order_id: str) -> Optional[CancelOrderResponse]:
        url = f"{self.config.account_rest_url}/orders/{order_id}/cancel"
        logger.info(f"cancel order: {order_id=}")

        res = await self._request_with_retry("PUT", url)
        if res is None:
            return None

        if res.status == 200:
            data = await res.json()
            logger.info(f"cancel limit order success: {data}")
            return parse_cancel_order_response(data)
        else:
            text = await res.text()
            logger.error(f"error canceling order: {res.status} {text}")
            return None



from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Optional

from ..helper import parse_time


# https://developer.oanda.com/rest-live-v20/instrument-df/#CandlestickGranularity
class CandlestickGranularity(Enum):
    M1 = "M1"  # 1 minute candlesticks, minute alignment
    # ...


# https://developer.oanda.com/rest-live-v20/instrument-df/#CandlestickData
@dataclass
class CandlestickData:
    o: Decimal
    h: Decimal
    l: Decimal
    c: Decimal


def parse_candlestick_data(data: dict) -> CandlestickData:
    return CandlestickData(
        o=Decimal(data["o"]),
        h=Decimal(data["h"]),
        l=Decimal(data["l"]),
        c=Decimal(data["c"]),
    )


# https://developer.oanda.com/rest-live-v20/instrument-df/#Candlestick
@dataclass
class Candlestick:
    time: datetime
    volume: int
    complete: bool
    bid: Optional[CandlestickData] = None
    ask: Optional[CandlestickData] = None
    mid: Optional[CandlestickData] = None


def parse_candlestick(data: dict) -> Candlestick:
    return Candlestick(
        time=parse_time(data["time"]),
        volume=data["volume"],
        complete=data["complete"],
        bid=parse_candlestick_data(data["bid"]) if "bid" in data else None,
        ask=parse_candlestick_data(data["ask"]) if "ask" in data else None,
        mid=parse_candlestick_data(data["mid"]) if "mid" in data else None,
    )



from datetime import datetime, timezone
from decimal import Decimal

from strats_oanda.model import (
    ClientPrice,
    CreateMarketOrderResponse,
    HomeConversionFactors,
    MarketOrderReason,
    MarketOrderTransaction,
    OrderFillReason,
    OrderFillTransaction,
    OrderPositionFill,
    PriceBucket,
    TimeInForce,
    TradeOpen,
    parse_create_market_order_response,
)


def test_parse_create_market_order_response():
    data = {
        "lastTransactionID": "79",
        "orderCreateTransaction": {
            "accountID": "101-009-31084545-001",
            "batchID": "78",
            "id": "78",
            "instrument": "USD_JPY",
            "positionFill": "DEFAULT",
            "reason": "CLIENT_ORDER",
            "requestID": "79368596638985858",
            "time": "2025-03-27T12:34:11.874521250Z",
            "timeInForce": "FOK",
            "type": "MARKET_ORDER",
            "units": "1",
            "userID": 31084545,
        },
        "orderFillTransaction": {
            "accountBalance": "3000000.0150",
            "accountID": "101-009-31084545-001",
            "baseFinancing": "0",
            "batchID": "78",
            "commission": "0.0000",
            "financing": "0.0000",
            "fullPrice": {
                "asks": [{"liquidity": "250000", "price": "150.763"}],
                "bids": [{"liquidity": "250000", "price": "150.759"}],
                "closeoutAsk": "150.769",
                "closeoutBid": "150.753",
                "timestamp": "2025-03-27T12:34:11.618910233Z",
            },
            "fullVWAP": "150.763",
            "gainQuoteHomeConversionFactor": "1",
            "guaranteedExecutionFee": "0.0000",
            "halfSpreadCost": "0.0020",
            "homeConversionFactors": {
                "gainBaseHome": {"factor": "150.459478"},
                "gainQuoteHome": {"factor": "1"},
                "lossBaseHome": {"factor": "151.062522"},
                "lossQuoteHome": {"factor": "1"},
            },
            "id": "79",
            "instrument": "USD_JPY",
            "lossQuoteHomeConversionFactor": "1",
            "orderID": "78",
            "pl": "0.0000",
            "price": "150.763",
            "quoteGuaranteedExecutionFee": "0",
            "quotePL": "0",
            "reason": "MARKET_ORDER",
            "requestID": "79368596638985858",
            "requestedUnits": "1",
            "time": "2025-03-27T12:34:11.874521250Z",
            "tradeOpened": {
                "guaranteedExecutionFee": "0.0000",
                "halfSpreadCost": "0.0020",
                "initialMarginRequired": "6.0304",
                "price": "150.763",
                "quoteGuaranteedExecutionFee": "0",
                "tradeID": "79",
                "units": "1",
            },
            "type": "ORDER_FILL",
            "units": "1",
            "userID": 31084545,
        },
        "relatedTransactionIDs": ["78", "79"],
    }
    got = parse_create_market_order_response(data)
    expect = CreateMarketOrderResponse(
        related_transaction_ids=["78", "79"],
        last_transaction_id="79",
        order_create_transaction=MarketOrderTransaction(
            id="78",
            time=datetime(2025, 3, 27, 12, 34, 11, 874521, tzinfo=timezone.utc),
            user_id=31084545,
            account_id="101-009-31084545-001",
            batch_id="78",
            type="MARKET_ORDER",
            request_id="79368596638985858",
            instrument="USD_JPY",
            units=Decimal("1"),
            time_in_force=TimeInForce.FOK,
            reason=MarketOrderReason.CLIENT_ORDER,
            price_bound=None,
            position_fill=OrderPositionFill.DEFAULT,
            trade_close=None,
            long_position_closeout=None,
            short_position_closeout=None,
            client_extensions=None,
            trade_client_extensions=None,
        ),
        order_fill_transaction=OrderFillTransaction(
            id="79",
            time=datetime(2025, 3, 27, 12, 34, 11, 874521, tzinfo=timezone.utc),
            user_id=31084545,
            account_id="101-009-31084545-001",
            batch_id="78",
            type="ORDER_FILL",
            request_id="79368596638985858",
            order_id="78",
            client_order_id=None,
            instrument="USD_JPY",
            units=Decimal("1"),
            home_conversion_factors=HomeConversionFactors(
                gain_quote_home={"factor": "1"},
                loss_quote_home={"factor": "1"},
                gain_base_home={"factor": "150.459478"},
                loss_base_home={"factor": "151.062522"},
            ),
            full_vwap=Decimal("150.763"),
            full_price=ClientPrice(
                type="PRICE",
                instrument=None,
                time=None,
                timestamp=datetime(2025, 3, 27, 12, 34, 11, 618910, tzinfo=timezone.utc),
                tradeable=None,
                bids=[PriceBucket(price=Decimal("150.759"), liquidity="250000")],
                asks=[PriceBucket(price=Decimal("150.763"), liquidity="250000")],
                closeout_bid=Decimal("150.753"),
                closeout_ask=Decimal("150.769"),
            ),
            reason=OrderFillReason.MARKET_ORDER,
            pl=Decimal("0.0000"),
            quote_pl=Decimal("0"),
            financing=Decimal("0.0000"),
            base_financing=Decimal("0"),
            quote_financing=None,
            commission=Decimal("0.0000"),
            guaranteed_execution_fee=Decimal("0.0000"),
            quote_guaranteed_execution_fee=Decimal("0"),
            account_balance=Decimal("3000000.0150"),
            trade_opened=TradeOpen(
                trade_id="79",
                units=Decimal("1"),
                price=Decimal("150.763"),
                half_spread_cost=Decimal("0.0020"),
                initial_margin_required=Decimal("6.0304"),
                client_extensions=None,
            ),
            trades_closed=None,
            trade_reduced=None,
            half_spread_cost=Decimal("0.0020"),
        ),
    )
    assert got == expect



from datetime import datetime, timezone
from decimal import Decimal

from strats_oanda.model import ClientPrice, PriceBucket, parse_client_price


def test_parse_client_price():
    data = {
        "type": "PRICE",
        "timestamp": "2025-03-24T15:34:25.366624289Z",
        "bids": [
            {"price": "150.693", "liquidity": 250000},
        ],
        "asks": [
            {"price": "150.697", "liquidity": 250000},
        ],
        "closeoutBid": "150.687",
        "closeoutAsk": "150.703",
        "status": "tradeable",
        "tradeable": True,
        "instrument": "USD_JPY",
    }
    expect = ClientPrice(
        type="PRICE",
        instrument="USD_JPY",
        time=None,
        timestamp=datetime(2025, 3, 24, 15, 34, 25, 366624, tzinfo=timezone.utc),
        tradeable=True,
        bids=[
            PriceBucket(
                price=Decimal("150.693"),
                liquidity=250000,
            ),
        ],
        asks=[
            PriceBucket(
                price=Decimal("150.697"),
                liquidity=250000,
            ),
        ],
        closeout_bid=Decimal("150.687"),
        closeout_ask=Decimal("150.703"),
    )
    assert parse_client_price(data) == expect


def test_parse_client_price_in_pricing_stream_api():
    data = {
        "type": "PRICE",
        "time": "2025-03-31T15:31:22.518120299Z",
        "bids": [
            {"price": "149.732", "liquidity": 250000},
        ],
        "asks": [
            {"price": "149.736", "liquidity": 250000},
        ],
        "closeoutBid": "149.727",
        "closeoutAsk": "149.742",
        "status": "tradeable",
        "tradeable": True,
        "instrument": "USD_JPY",
    }
    expect = ClientPrice(
        type="PRICE",
        instrument="USD_JPY",
        time=datetime(2025, 3, 31, 15, 31, 22, 518120, tzinfo=timezone.utc),
        timestamp=None,
        tradeable=True,
        bids=[
            PriceBucket(
                price=Decimal("149.732"),
                liquidity=250000,
            ),
        ],
        asks=[
            PriceBucket(
                price=Decimal("149.736"),
                liquidity=250000,
            ),
        ],
        closeout_bid=Decimal("149.727"),
        closeout_ask=Decimal("149.742"),
    )
    assert parse_client_price(data) == expect
